local nvim_lsp   = require 'lspconfig'
local protocol   = require 'vim.lsp.protocol'
local ts_utils   = require 'nvim-lsp-ts-utils'
local util       = require 'lspconfig/util'
local lightbulb  = require 'nvim-lightbulb'

-- https://github.com/neovim/nvim-lspconfig/wiki/User-contributed-tips
function FormatRangeOperator()
  local old_func = vim.go.operatorfunc
  _G.op_func_formatting = function()
    local start = vim.api.nvim_buf_get_mark(0, '[')
    local finish = vim.api.nvim_buf_get_mark(0, ']')
    vim.lsp.buf.range_formatting({}, start, finish)
    vim.go.operatorfunc = old_func
    _G.op_func_formatting = nil
  end
  vim.go.operatorfunc = 'v:lua.op_func_formatting'
  vim.api.nvim_feedkeys('g@', 'n', false)
end

local function preview_location_callback(_, result)
  if result == nil or vim.tbl_isempty(result) then
    return nil
  end
  vim.lsp.util.preview_location(result[1])
end

function PeekDefinition()
  local params = vim.lsp.util.make_position_params()
  return vim.lsp.buf_request(0, 'textDocument/definition', params, preview_location_callback)
end

protocol.CompletionItemKind = {
  ' Text'         ;
  ' Method'       ;
  ' Function'     ;
  '全Constructor'  ;
  ' Field'        ;
  ' Variable'     ;
  ' Class'        ;
  ' Interface'    ;
  ' Module'       ;
  ' Property'     ;
  ' Unit'         ;
  ' Value'        ;
  ' Enum'         ;
  ' Keyword'      ;
  ' Snippet'      ;
  ' Color'        ;
  ' File'         ;
  ' Reference'    ;
  ' Folder'       ;
  ' EnumMember'   ;
  ' Constant'     ;
  ' Struct'       ;
  ' Event'        ;
  ' Operator'     ;
  ' TypeParameter';
}

local ts_utils_setup = function(client, bufnr, opts)
  if client.name == 'tsserver' then
    -- Disable tsserver formatting, we want formatting via prettier
    client.resolved_capabilities.document_formatting       = false
    client.resolved_capabilities.document_range_formatting = false

    ts_utils.setup {
      debug                          = false,
      disable_commands               = false,
      enable_import_on_completion    = false,

      import_all_timeout             = 5000,
      import_all_scan_buffers        = 100,
      import_all_select_source       = false,
      import_all_priorities = {
        buffers        = 4, -- loaded buffer names
        buffer_content = 3, -- loaded buffer content
        local_files    = 2, -- git files or files with relative path markers
        same_file      = 1, -- add to existing import statement
      },

      update_imports_on_move         = false,
      require_confirmation_on_move   = false,
      watch_dir                      = nil,

      filter_out_diagnostics_by_severity = {},
      filter_out_diagnostics_by_code     = {},

      auto_inlay_hints      = true,
      inlay_hints_highlight = "Hint",
    }

    ts_utils.setup_client(client)

    vim.api.nvim_buf_set_keymap(bufnr, 'n', 'go', ':TSLspOrganize<CR>'  , opts)
    vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gf', ':TSLspFixCurrent<CR>', opts)
    vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gI', ':TSLspImportAll<CR>' , opts)
  end
end

local on_attach = function(client, bufnr)
  vim.api.nvim_buf_set_option(bufnr, 'omnifunc', 'v:lua.vim.lsp.omnifunc')
  vim.lsp.set_log_level('warn')

  if client.config.flags then
    client.config.flags.allow_incremental_sync = true
    client.config.flags.debounce_text_changes  = 100
  end

  local opts = { noremap=true, silent=true }

  -- This needs to be here, so we disable formatting with tsserver before
  -- actually checking it below.
  ts_utils_setup(client, bufnr, opts)

  vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gd'       , '<cmd>lua vim.lsp.buf.definition()<CR>'       , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gD'       , '<cmd>lua vim.lsp.buf.declaration()<CR>'      , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gr'       , '<cmd>lua vim.lsp.buf.references()<CR>'       , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gi'       , '<cmd>lua vim.lsp.buf.implementation()<CR>'   , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', '<C-k>'    , '<cmd>lua vim.lsp.buf.type_definition()<CR>'  , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', '1gd'      , '<cmd>lua vim.lsp.buf.document_symbol()<CR>'  , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', '1gD'      , '<cmd>lua vim.lsp.buf.workspace_symbol()<CR>' , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gR'       , '<cmd>lua vim.lsp.buf.rename()<CR>'           , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', 'pd'       , '<cmd>lua PeekDefinition()<CR>'               , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', '<Leader>k', '<cmd>lua vim.lsp.buf.hover()<CR>'            , opts)
  vim.api.nvim_buf_set_keymap(bufnr, 'n', '<Leader>S', '<cmd>lua vim.lsp.buf.signature_help()<CR>'   , opts)

  if client.resolved_capabilities.code_action then
    vim.api.nvim_buf_set_keymap(bufnr, 'n', 'ga'       , '<cmd>lua vim.lsp.buf.code_action()<CR>'      , opts)
    vim.api.nvim_buf_set_keymap(bufnr, 'v', 'ga'       , '<cmd>lua vim.lsp.buf.range_code_action()<CR>', opts)
    vim.api.nvim_buf_set_keymap(bufnr, 'n', '<Leader>r', '<cmd>lua vim.lsp.buf.code_action{ only = \'refactor\' }<CR>'      , opts)
    vim.api.nvim_buf_set_keymap(bufnr, 'v', '<Leader>r', '<cmd>lua vim.lsp.buf.range_code_action{ only = \'refactor\' }<CR>', opts)
  end

  if client.resolved_capabilities.document_range_formatting then
    vim.api.nvim_buf_set_keymap(bufnr, 'x', 'gq', '<cmd>lua FormatRangeOperator()<CR>', opts)
  end
  if client.resolved_capabilities.document_formatting then
    vim.api.nvim_buf_set_keymap(bufnr, 'n', 'gq', '<cmd>lua vim.lsp.buf.formatting()<CR>', opts)
  end

  if client.resolved_capabilities.code_lens or client.resolved_capabilities.document_highlight then
    vim.api.nvim_create_augroup("LSP", {clear = true})
  end

  if client.resolved_capabilities.code_lens then
    vim.api.nvim_buf_set_keymap(bufnr, "n", "<Leader>l", "<cmd>lua vim.lsp.codelens.run()<CR>", opts)
    vim.api.nvim_create_autocmd({"CursorHold", "CursorHoldI", "InsertLeave"}, {
      group   = "LSP",
      buffer  = bufnr,
      callback = function ()
        vim.lsp.codelens.refresh()
      end,
    })
    vim.api.nvim_create_autocmd({"CursorHold", "CursorHoldI"}, {
      group   = "LSP",
      buffer  = bufnr,
      callback = function ()
        lightbulb.update_lightbulb()
      end,
    })
  end

  if client.resolved_capabilities.document_highlight then
    vim.api.nvim_create_autocmd("CursorHold", {
      group   = "LSP",
      buffer  = bufnr,
      callback = function ()
        vim.lsp.buf.document_highlight()
      end,
    })
    vim.api.nvim_create_autocmd("CursorMoved", {
      group   = "LSP",
      buffer  = bufnr,
      callback = function ()
        vim.lsp.buf.clear_references()
      end,
    })
  end
end

local on_init = function(client)
  if client.config.settings then
    client.notify('workspace/didChangeConfiguration', { settings = client.config.settings })
  end
end

local capabilities = vim.lsp.protocol.make_client_capabilities()
capabilities.workspace.configuration = true
capabilities.textDocument.completion.completionItem.snippetSupport = true

local rust_tool_opts = {
  tools = {
    autoSetHints       = true,
    hover_with_actions = false,
    runnables          = { use_telescope = false },
    debuggables        = { use_telescope = false },
    inlay_hints = {
      only_current_line         = false,
      only_current_line_autocmd = "CursorHold",
      show_parameter_hints      = true,
      parameter_hints_prefix    = "<- ",
      other_hints_prefix        = "=> ",
      highlight                 = "Hint",
    },
    hover_actions = {
      auto_focus = true
    },
    crate_graph = {
      backend = "svg",
      output  = "crates_graph.svg",
      full    = false,
    }
  },
  server = {
    on_init      = on_init,
    on_attach    = on_attach,
    capabilities = capabilities,
    settings = {
      ["rust-analyzer"] = {
        checkOnSave = {
          command = "clippy"
        },
      }
    }
  },
}

nvim_lsp.clangd.setup {
  on_init      = on_init,
  on_attach    = on_attach,
  capabilities = capabilities,
  default_config = {
    cmd       = { "clangd", "--background-index", "--pch-storage=memory", "--clang-tidy", "--suggest-missing-includes", "--header-insertion=never" },
    filetypes = { 'c', 'cpp' },
    root_dir = function(fname)
      -- We specify build/compile_commands.json as that is where the compile_commands.json
      -- gets generated automatically for meson projects.
      local root_pattern = util.root_pattern('build/compile_commands.json', 'compile_commands.json', 'compile_flags.txt', '.git')
      local filename     = util.path.is_absolute(fname) and fname or util.path.join(vim.loop.cwd(), fname)
      return root_pattern(filename) or util.path.dirname(filename)
    end,
  }
}

nvim_lsp.hls.setup {
  on_init      = on_init,
  on_attach    = on_attach,
  capabilities = capabilities
}

nvim_lsp.jedi_language_server.setup {
  on_init      = on_init,
  on_attach    = on_attach,
  capabilities = capabilities,
  init_options = {
    completion = {
      resolveEagerly = true,
    },
    jediSettings = {
      caseInsensitiveCompletion = false,
    },
    workspace = {
      symbols = {
        maxSymbols = 50
      },
    },
  },
}

require('rust-tools').setup(rust_tool_opts)

nvim_lsp.tsserver.setup {
  init_options = require("nvim-lsp-ts-utils").init_options,
  on_init      = on_init,
  on_attach    = on_attach,
  capabilities = capabilities,
}

nvim_lsp.eslint.setup {
  on_init      = on_init,
  on_attach    = on_attach,
  capabilities = capabilities
}
